#!/usr/bin/env python
# -*- mode: python; coding: koi8-r; -*-
# (c) Con Radchenko mailto:bofh@pochta.ru

import sys, os
from glob import glob
import time
from ConfigParser import ConfigParser

import gtk
from gtk import TRUE, FALSE
from gtk import glade
from gtk import gdk
#import pango

import ornamentbook
from miscutils import entry_dialog, error_dialog, fix_filename, OpenFileDialog
from properties import Properties, PropertiesDialog


glade_file = 'ornamentbook.glade'
propsdialog_glade_file = 'propsdialog.glade'
config_dir = os.path.expanduser('~/.pybr')
config_file = os.path.expanduser('~/.pybr/ornamentbook.conf')
user_skin_dir = os.path.expanduser('~/.pybr/skins')

##----------------------------------------------------------------------

class NewPyBR:

    def __init__(self, filename=None, fullscreen=False):

        self.filename = None
        self.encoding = 'utf-8'

        self.is_fullscreen = False

        self.properties = Properties()
        #self.properties.read_properties(config_file)

        self.create_widgets()

        if filename:
            gtk.idle_add(self.open_file, filename)
        elif self.properties.recent_files:
            gtk.idle_add(self.open_file, self.properties.recent_files[-1][0])

        if fullscreen and gtk.pygtk_version >= (2, 2):
            self.on_fullscreen_activate()

        OpenFileDialog.file_select_path = self.properties.file_select_path

        #self.timeout_cb()
        gtk.idle_add(lambda : self.timeout_cb() and False)
        gtk.timeout_add(10000, self.timeout_cb)


    def create_widgets(self):

        gf = None
        for d in sys.path:
            f = os.path.join(d, 'ornamentbook', glade_file)
            if os.path.exists(f):
                gf = f
                break
        if not gf:
            sys.exit('ERROR: can\'t find glade file')
        self.glade_file = gf

        pgf = None
        for d in sys.path:
            f = os.path.join(d, 'ornamentbook', propsdialog_glade_file)
            if os.path.exists(f):
                pgf = f
                break
        if not pgf:
            sys.exit('ERROR: can\'t find glade file')
        self.propsdialog_glade_file = pgf

        wTree = glade.XML(gf, 'main_window')
        wTree.signal_autoconnect(self)

        self.window = wTree.get_widget('main_window')
        top_vbox = wTree.get_widget('top_vbox')
        text_widget = ornamentbook.OrnamentBook(self.window, self.properties)
        top_vbox.pack_start(text_widget)

        text_widget.realize()
        #text_widget.connect('position-changed', self.position_changed_cb)
        self.text_widget = text_widget

        self.main_menubar = wTree.get_widget('main_menubar')
        # (un)fullscreen available in PyGTK 2.2 and above
        if gtk.pygtk_version < (2, 2):
            wTree.get_widget('fullscreen_menu').set_sensitive(False)

        self.open_recent_file_menu = wTree.get_widget('open_recent_file_menu')
        self.skin_menu = wTree.get_widget('skin_menu')
        #self.percent_label = wTree.get_widget('percent_label')
        self.hyphenation_menu = wTree.get_widget('hyphenation_menu')

        self.properties.read_properties(config_file)
        self.window.resize(self.properties.main_window_width, self.properties.main_window_height)


        self.update_skin_menu()
        self.update_hyphenation_menu()


        self.window.show_all()
        text_widget.initialize()


        #text_widget.grab_focus()


##     def position_changed_cb(self, w, pos):
##         self.percent_label.set_text('%d %%' % int(pos*100))


    def open_file(self, filename):
        #print 'open_file', filename

        try:
            open(filename)
        except IOError, err:
            #sys.stderr.write('ERROR: can\'t open file: %s: %s\n'
            #                 % (filename, err[1]))
            error_dialog(self.window, "Can't open file %s: %s" %
                         (fix_filename(filename), err[1]))
            self.update_recent_files_menu()
            return

        if self.filename: # save previous file
            #print 'save previous file', self.text_widget.current_page_line_index
            self.add_file_to_recent(self.filename,
                                    self.text_widget.current_page_line_index,
                                    self.text_widget.current_page_char_index,
                                    self.text_widget.encoding,
                                    self.text_widget.lang)


        rf = self.get_recent_file(filename)

        if rf:
            ret = self.text_widget.open_file(filename,
                                             rf[3], rf[1], rf[2], rf[4])
        else:
            ret = self.text_widget.open_file(filename)

        if ret: # success
            self.filename = filename
            self.add_file_to_recent(self.filename,
                                    self.text_widget.current_page_line_index,
                                    self.text_widget.current_page_char_index,
                                    self.text_widget.encoding,
                                    self.text_widget.lang)
            self.update_recent_files_menu()


    def timeout_cb(self):

        title = self.properties.title_format
        title = title.replace('%f', '%(filename)s')
        title = title.replace('%p', '%(position)d%%')
        title = title.replace('%t', '%(time)s')

        try:
            title = title % {'filename': fix_filename(self.filename),
                             'position': int(self.text_widget.calc_position()*100),
                             'time': time.strftime('%H:%M', time.localtime()),
                             }
        except:
            title = 'OrnamentBook'
        self.window.set_title(title)

        self.text_widget.draw_status()

        return True


    def update_recent_files_menu(self):
        menu = gtk.Menu()
        menu.show()
        for f in self.properties.recent_files:
            menuitem = gtk.MenuItem(fix_filename(f[0]))
            menuitem.get_child().set_use_underline(FALSE)
            menuitem.connect('activate', self.recent_files_activate_cb, f[0])
            menu.add(menuitem)
            menuitem.show()
        self.open_recent_file_menu.set_submenu(menu)


    def update_hyphenation_menu(self):
        menu = gtk.Menu()
        menu.show()
        from hyphenation import Hyphenation
        langs = Hyphenation().get_langs()
        for l, n in langs:
            menuitem = gtk.MenuItem(n)
            menuitem.get_child().set_use_underline(FALSE)
            menuitem.connect('activate', self.hyphenation_menu_activate_cb, l)
            menu.add(menuitem)
            menuitem.show()
        self.hyphenation_menu.set_submenu(menu)


    def update_skin_menu(self):

        skins = []
        if os.path.exists(user_skin_dir) and os.path.isdir(user_skin_dir):
            skins_dirs = [user_skin_dir]
        else:
            skins_dirs = []

        for d in sys.path:
            f = os.path.join(d, 'ornamentbook', ornamentbook.skins_dir)
            if os.path.exists(f) and os.path.isdir(f):
                skins_dirs.append(f)
                break

        for d in skins_dirs:
            f = os.path.join(d, '*.jpg')
            skins.extend(glob(f))

        menu = gtk.Menu()
        menu.show()
        for f in skins:
            fn = os.path.abspath(f)
            menuitem = gtk.MenuItem(fn)
            menuitem.get_child().set_use_underline(FALSE)
            menuitem.connect('activate', self.skin_menu_activate_cb, fn)
            menu.add(menuitem)
            menuitem.show()
        self.skin_menu.set_submenu(menu)


    def skin_menu_activate_cb(self, w, filename):
        #print 'skin_menu_activate_cb', filename
        self.text_widget.set_skin(filename)
        self.properties.path_to_skin = filename
        self.text_widget.draw_text()


    def hyphenation_menu_activate_cb(self, w, lang):
        self.text_widget.lang = lang
        self.text_widget.draw_text()


    def recent_files_activate_cb(self, w, filename):
        #print 'recent_files_activate_cb', filename
        self.open_file(filename)


    def get_recent_file(self, filename):
        #print 'get_recent_file', filename
        for f in self.properties.recent_files:
            if f[0] == filename:
                return f
        return None


    def add_file_to_recent(self, filename, line, index, encoding, lang):

        i = 0
        for f in self.properties.recent_files:
            if f[0] == filename:
                del self.properties.recent_files[i]
                break
            i+=1
        if len(self.properties.recent_files) >= self.properties.num_recent_files:
            del self.properties.recent_files[:len(self.properties.recent_files)-self.properties.num_recent_files+1]
        self.properties.recent_files.append((filename, line, index, encoding, lang))

        #self.update_recent_files_menu()


    def on_open_activate(self, *args):

        dialog = OpenFileDialog(self.window)
        filename = dialog.filename
        if filename:
            self.open_file(filename)
        return


    def on_go_to_position_activate(self, *args):
        s = entry_dialog(self.window, 'Go to position', 'Go to position: (%) ')
        try:
            pos = float(s)
        except:
            return
        if 100 < pos < 0:
            return
        self.text_widget.set_position(pos/100)


    def on_quit_activate(self, *args):
        self.window.destroy()


    def on_main_window_unrealize(self, *args):
        #print 'on_main_window_unrealize'
        self.properties.main_window_width, self.properties.main_window_height = self.window.get_size()
        self.properties.file_select_path = OpenFileDialog.file_select_path

        if self.filename:
            self.add_file_to_recent(self.filename,
                                    self.text_widget.current_page_line_index,
                                    self.text_widget.current_page_char_index,
                                    self.text_widget.encoding,
                                    self.text_widget.lang)
        self.properties.save_properties(config_file)


    def on_properties_activate(self, *args):
        dlg = PropertiesDialog(self.window, self.properties, self.propsdialog_glade_file)
        dlg.connect('properties-apply', self.apply_properties_cb, dlg)


    def apply_properties_cb(self, w, dlg):
        #print 'apply_properties_cb'
        self.properties = dlg.properties
        self.text_widget.properties = dlg.properties
        self.text_widget.initialize()
        self.timeout_cb()
        self.text_widget.draw_text()


    def on_increase_font_activate(self, *args):
        self.text_widget.incr_font(1)
        self.text_widget.draw_text()


    def on_decrease_font_activate(self, *args):
        self.text_widget.incr_font(-1)
        self.text_widget.draw_text()


    def on_fullscreen_activate(self, *args):

        if self.is_fullscreen:
            self.main_menubar.show()
            self.window.unfullscreen()
        else:
            if self.properties.hide_menu:
                self.main_menubar.hide()
            self.window.fullscreen()
        self.is_fullscreen = not self.is_fullscreen

##----------------------------------------------------------------------

if __name__ == '__main__':
##     try:
##         os.mkdir(config_dir)
##     except:
##         pass


    npbr = NewPyBR()
    npbr.window.connect('destroy', lambda w: gtk.main_quit())
    gtk.main()


